use std::{env, io, process};
use std::fs::File;
use std::io::{BufRead, BufReader};
use crate::CifError::{CapacityError, GrammarError, InvalidHeaderError, IoError};
use arrayvec::{ArrayString, ArrayVec};
use thiserror::Error;

mod nummap;

#[cfg(test)]
mod test;

#[derive(Error, Debug)]
enum CifError {
    #[error("The number string was too long")]
    CapacityError,
    #[error("Number is invalid")]
    GrammarError,
    #[error("An IO error has occurred: {0}")]
    IoError(io::Error),
    #[error("CIF header is invalid")]
    InvalidHeaderError,
}

impl<T> From<arrayvec::CapacityError<T>> for CifError {
    fn from(_: arrayvec::CapacityError<T>) -> Self {
        CapacityError
    }
}

impl From<io::Error> for CifError {
    fn from(e: io::Error) -> Self {
        IoError(e)
    }
}

type Result<T> = std::result::Result<T, CifError>;

#[inline]
fn join_to_arraystring<const C: usize>(elems: &[&str], str: &mut ArrayString<C>) -> Result<()> {
    for (i, item) in elems.iter().enumerate() {
        if i > 0 {
            str.try_push(' ')?;
        }
        str.try_push_str(item)?;
    }
    Ok(())
}

fn unfuck_cif_number(number: &str) -> Result<u32> {
    // jeśli ktoś zdoła przekroczyć te bufory, to jest oficjalnie pojebany
    let mut buf = ArrayString::<2048>::new();
    let split = number.split_whitespace().collect::<ArrayVec<_, 64>>();
    let splitlen = split.len();
    for (i, word) in split.iter().enumerate() {
        if !word.starts_with("ty") {
            continue;
        }
        let thousands = if *word == "tysiąc" {
            if i == 0 {
                &1_u16
            } else {
                return Err(GrammarError);
            }
        } else if *word == "tysięcy" || *word == "tysiące" {
            if i > 0 {
                join_to_arraystring(&split[..i], &mut buf)?;
                nummap::NUMMAP.get(&buf).ok_or(GrammarError)?
            } else {
                return Err(GrammarError);
            }
        } else {
            return Err(GrammarError);
        };
        let remainder = if i + 1 < splitlen {
            buf.clear();
            join_to_arraystring(&split[i + 1..], &mut buf)?;
            nummap::NUMMAP.get(&buf).ok_or(GrammarError)?
        } else {
            &0_u16
        };
        return Ok(*thousands as u32 * 1000 + *remainder as u32);
    }
    Ok(*nummap::NUMMAP.get(number).ok_or(GrammarError)? as u32)
}

// ujebuje stringa na dwie części po znalezieniu spacji
#[inline]
fn separate_string(str: &str) -> Option<(&str, &str)> {
    if let Some(o) = str.find(' ') {
        Some((&str[..o], &str[o + 1..]))
    } else {
        None
    }
}

// to samo co wyżej ale tylko od 0 do 999
#[inline]
fn unfuck_cif_number_fast(number: &str) -> Result<u16> {
    Ok(*(nummap::NUMMAP.get(number).ok_or(GrammarError)?))
}

fn decoder(inputfile: &str, outputfile: &str) -> Result<()> {
    let infile = File::open(inputfile)?;
    let mut lines = BufReader::new(infile).lines();

    // musi być "CIF: " na początku. Na razie wyjebane we flagi
    if !lines.next().transpose()?.ok_or(InvalidHeaderError)?.starts_with("CIF: ") {
        return Err(InvalidHeaderError)
    }

    // wspieramy tylko wersję 1 CIFa
    if lines.next().transpose()?.ok_or(InvalidHeaderError)? != "WERSJA jeden" {
        return Err(InvalidHeaderError)
    }

    let size_string = lines.next().transpose()?.ok_or(InvalidHeaderError)?;
    if size_string.len() < 9 || !size_string.starts_with("ROZMIAR ") {
        return Err(InvalidHeaderError)
    }
    let size_split = (&size_string[8..]).split(", ").take(3).collect::<ArrayVec<_, 3>>();
    if size_split.len() != 3 || size_split[0] != "szerokość:" || size_split[1] != "wysokość:" || size_split[2] != "bitów_na_piksel" {
        return Err(InvalidHeaderError)
    }


    Ok(())
}

fn main() {
    let argv = env::args().skip(1).take(2).collect::<ArrayVec<_, 2>>();
    if argv.len() != 2 {
        eprintln!("Invalid usage;\nUsage: zardzewialy-dekoder-cif <input cif file> <output bmp file>");
        process::exit(1);
    }
    if let Err(e) = decoder(&argv[0], &argv[1]) {
        eprintln!("Error: {}", e);
        process::exit(1);
    }
}
